module GridMerger

  #Return the face group and the edge group
  def self.parse_input(sel)
    gs = sel.grep(Sketchup::Group)
    return nil, nil if gs.length != 2
    fg = gs.select { |g| g.entities.grep(Sketchup::Face).length > 0}
    return nil, nil if fg.length < 1 || fg.length > 1
    eg = gs.select { |g| g.entities.grep(Sketchup::Face).length == 0}
    return nil, nil if eg.length < 1 || eg.length > 1
    return fg[0], eg[0]
  end
  
  #organize edges into classes based on direction
  def self.organize_edges(eg)
    es = eg.entities.grep(Sketchup::Edge)
    classes = [] #an array of vectors, where a single vector defines a class
    ess = [] #array of arrays where every sub array contains edges in a specific class
    es.each { |e|
      v0 = (e.end.position - e.start.position)
      cl = compute_class(v0, classes)
      if cl == -1  #create a new class
        classes << v0
        ess << [e]
      else
        ess[cl] << e
      end
    }
    return ess
  end
  
  def self.compute_class(v0, classes)
    (0..classes.length - 1).each { |i|
      cv = classes[i]
      return i if (cv.angle_between(v0) < Math::PI / 8) ||
                  (cv.angle_between(v0.reverse) < Math::PI / 8)
    }
    return -1
  end
  
  def self.merge(fg, eg, ess)
  
    max_edges_per_group = 250
    tr = eg.transformation * fg.transformation.inverse
    sgs = [] 
    (0..ess.length - 1).each { |i|
      count = 0
      #construct groups for each class depending on the number of edges;
      #many small explosions are faster than one big (within limits).
      sgs << []
      no_of_subgroups = [(ess[i].length / max_edges_per_group), 1].max
      (1..no_of_subgroups).each { |j| sgs[i] << fg.entities.add_group }
      ess[i].each { |e|
        sgs[i][count % sgs[i].length].entities.add_line(e.start.position.transform(tr), e.end.position.transform(tr))
        count += 1
      }
    }
    
    exps = 0
    sgs.each { |gs| gs.each { |g| g.explode; exps += 1 }}
    
    puts 'Explosions: '+exps.to_s
    eg.entities.clear!
  
  end
  
  ################################
  ######## VERIFICATION ##########
  ################################
  
  def self.verify(fg)
    bow_ties = find_bow_ties(fg)
    no_face_edges = find_no_face_edges(fg)
    puts '--== VERIFICATION ==--'
    puts 'Bow ties:  '+bow_ties.length.to_s
    puts 'No-face-edges: '+no_face_edges.length.to_s
    puts '----------------------'
  end
  
  #a face spreading over many cells leaves edges with no face...
  def self.find_no_face_edges(fg)
    es = fg.entities.grep(Sketchup::Edge)
    return es.select { |e| e.faces.length == 0 }
  end
  
  #A bow tie is defined by having the same vertex appear many times in the loop
  def self.find_bow_ties(fg)
    fs = fg.entities.grep(Sketchup::Face)
    return fs.select { |f|
      f.outer_loop.vertices.length != f.outer_loop.vertices.uniq.length
    }
  end
  
  ################################
  
  def self.main
    mod = Sketchup.active_model # Open model
    ent = mod.entities # All entities in model
    sel = mod.selection # Current selection
    
    t0 = Time.now
    mod.start_operation("GridMerge", true)
    fg, eg = parse_input(sel)
    if fg == nil
      puts 'Input must be a group with face(s) and a group with only edges'
      return
    end
    ess = organize_edges(eg)
    merge(fg, eg, ess)
    verify(fg)
    mod.commit_operation()
    t1 = Time.now
    puts 'Total Time: '+(t1 - t0).to_s+' s'
  end
  
  main
  
end
